/*
 * Jalview - A Sequence Alignment Editor and Viewer ($$Version-Rel$$)
 * Copyright (C) $$Year-Rel$$ The Jalview Authors
 * 
 * This file is part of Jalview.
 * 
 * Jalview is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation, either version 3
 * of the License, or (at your option) any later version.
 *  
 * Jalview is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty 
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
 * PURPOSE.  See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jalview.  If not, see <http://www.gnu.org/licenses/>.
 * The Jalview Authors are detailed in the 'AUTHORS' file.
 */
package jalview.gui;

import jalview.util.MessageManager;
import jalview.util.Platform;
import jalview.viewmodel.OverviewDimensions;
import jalview.viewmodel.OverviewDimensionsHideHidden;
import jalview.viewmodel.OverviewDimensionsShowHidden;
import jalview.viewmodel.ViewportListenerI;

import java.awt.BorderLayout;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.beans.PropertyChangeEvent;

import javax.swing.JCheckBoxMenuItem;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;

/**
 * Panel displaying an overview of the full alignment, with an interactive box
 * representing the viewport onto the alignment.
 * 
 * @author $author$
 * @version $Revision$
 */
public class OverviewPanel extends JPanel
        implements Runnable, ViewportListenerI
{
  private OverviewDimensions od;

  private OverviewCanvas oviewCanvas;

  private AlignViewport av;

  private AlignmentPanel ap;

  private JCheckBoxMenuItem displayToggle;

  private boolean showHidden = true;

  private boolean draggingBox = false;

  /**
   * Creates a new OverviewPanel object.
   * 
   * @param alPanel
   *          The alignment panel which is shown in the overview panel
   */
  public OverviewPanel(AlignmentPanel alPanel)
  {
    this.av = alPanel.av;
    this.ap = alPanel;

    od = new OverviewDimensionsShowHidden(av.getRanges(),
            (av.isShowAnnotation()
                    && av.getAlignmentConservationAnnotation() != null));

    setSize(od.getWidth(), od.getHeight());

    oviewCanvas = new OverviewCanvas(od, av);
    setLayout(new BorderLayout());
    add(oviewCanvas, BorderLayout.CENTER);

    av.getRanges().addPropertyChangeListener(this);

    addComponentListener(new ComponentAdapter()
    {
      @Override
      public void componentResized(ComponentEvent evt)
      {
        if ((getWidth() != od.getWidth())
                || (getHeight() != (od.getHeight())))
        {
          updateOverviewImage();
          setBoxPosition();
        }
      }
    });

    addMouseMotionListener(new MouseMotionAdapter()
    {
      @Override
      public void mouseDragged(MouseEvent evt)
      {
        if (!SwingUtilities.isRightMouseButton(evt))
        {
          if (draggingBox)
          {
            // set the mouse position as a fixed point in the box
            // and drag relative to that position
            od.adjustViewportFromMouse(evt.getX(), evt.getY(),
                    av.getAlignment().getHiddenSequences(),
                    av.getAlignment().getHiddenColumns());
          }
          else
          {
            od.updateViewportFromMouse(evt.getX(), evt.getY(),
                    av.getAlignment().getHiddenSequences(),
                    av.getAlignment().getHiddenColumns());
          }
        }
      }

      @Override
      public void mouseMoved(MouseEvent evt)
      {
        if (od.isPositionInBox(evt.getX(), evt.getY()))
        {
          // display drag cursor at mouse position
          setCursor(Cursor.getPredefinedCursor(Cursor.MOVE_CURSOR));
        }
        else
        {
          // reset cursor
          setCursor(Cursor.getPredefinedCursor(Cursor.DEFAULT_CURSOR));
        }
      }
    });

    addMouseListener(new MouseAdapter()
    {
      @Override
      public void mousePressed(MouseEvent evt)
      {
        if (SwingUtilities.isRightMouseButton(evt))
        {
          if (!Platform.isAMac())
          {
            showPopupMenu(evt);
          }
        }
        else
        {
          // don't do anything if the mouse press is in the overview's box
          // (wait to see if it's a drag instead)
          // otherwise update the viewport
          if (!od.isPositionInBox(evt.getX(), evt.getY()))
          {
            draggingBox = false;
            od.updateViewportFromMouse(evt.getX(), evt.getY(),
                    av.getAlignment().getHiddenSequences(),
                    av.getAlignment().getHiddenColumns());
          }
          else
          {
            draggingBox = true;
            od.setDragPoint(evt.getX(), evt.getY(),
                    av.getAlignment().getHiddenSequences(),
                    av.getAlignment().getHiddenColumns());
          }
        }
      }

      @Override
      public void mouseClicked(MouseEvent evt)
      {
        if (SwingUtilities.isRightMouseButton(evt))
        {
          showPopupMenu(evt);
        }
      }
    });

    updateOverviewImage();
  }

  /*
   * Displays the popup menu and acts on user input
   */
  private void showPopupMenu(MouseEvent e)
  {
    JPopupMenu popup = new JPopupMenu();
    ActionListener menuListener = new ActionListener()
    {
      @Override
      public void actionPerformed(ActionEvent event)
      {
        // switch on/off the hidden columns view
        toggleHiddenColumns();
        displayToggle.setSelected(showHidden);
      }
    };
    displayToggle = new JCheckBoxMenuItem(
            MessageManager.getString("label.togglehidden"));
    displayToggle.setEnabled(true);
    displayToggle.setSelected(showHidden);
    popup.add(displayToggle);
    displayToggle.addActionListener(menuListener);
    popup.show(this, e.getX(), e.getY());
  }

  /*
   * Toggle overview display between showing hidden columns and hiding hidden columns
   */
  private void toggleHiddenColumns()
  {
    if (showHidden)
    {
      showHidden = false;
      od = new OverviewDimensionsHideHidden(av.getRanges(),
              (av.isShowAnnotation()
                      && av.getAlignmentConservationAnnotation() != null));
    }
    else
    {
      showHidden = true;
      od = new OverviewDimensionsShowHidden(av.getRanges(),
              (av.isShowAnnotation()
                      && av.getAlignmentConservationAnnotation() != null));
    }
    oviewCanvas.resetOviewDims(od);
    updateOverviewImage();
    setBoxPosition();
  }

  /**
   * Updates the overview image when the related alignment panel is updated
   */
  public void updateOverviewImage()
  {
    if (oviewCanvas == null)
    {
      /*
       * panel has been disposed
       */
      return;
    }

    if ((getWidth() > 0) && (getHeight() > 0))
    {
      od.setWidth(getWidth());
      od.setHeight(getHeight());
    }

    setPreferredSize(new Dimension(od.getWidth(), od.getHeight()));

    if (oviewCanvas.restartDraw())
    {
      return;
    }

    Thread thread = new Thread(this);
    thread.start();
    repaint();
  }

  @Override
  public void run()
  {
    oviewCanvas.draw(av.isShowSequenceFeatures(),
            (av.isShowAnnotation()
                    && av.getAlignmentConservationAnnotation() != null),
            ap.getSeqPanel().seqCanvas.getFeatureRenderer());
    setBoxPosition();
  }

  /**
   * Update the overview panel box when the associated alignment panel is
   * changed
   * 
   */
  private void setBoxPosition()
  {
    od.setBoxPosition(av.getAlignment().getHiddenSequences(),
            av.getAlignment().getHiddenColumns());
    repaint();
  }

  @Override
  public void propertyChange(PropertyChangeEvent evt)
  {
    setBoxPosition();
  }

  /**
   * Removes this object as a property change listener, and nulls references
   */
  protected void dispose()
  {
    try
    {
      av.getRanges().removePropertyChangeListener(this);
    } finally
    {
      av = null;
      oviewCanvas = null;
      ap = null;
      od = null;
    }
  }
}
